#!/usr/bin/env python

import numpy as np

from contspy import Continuation, plot_continuation_results


class Coupled(Continuation):
    def __init__(self, L, N, nvar):
        Continuation.__init__(self, L, N, nvar)
        self.Ar = 5.0

    def Res(self, u, lmbda):
        """
      The residual to solve for the system
    """
        res0 = self.Res0(u, lmbda)
        res1 = self.Res1(u, lmbda)

        return np.hstack((res0, res1))

    def Res0(self, u, lmbda):
        """
        Same residual as in test_S_coupled.py
    """
        u0, _ = np.split(u, 2)
        return np.dot(self.D2, u0) + lmbda * np.exp(np.divide(self.Ar * u0, 1.0 + u0))

    def Res1(self, u, lmbda):
        """
        Only diffusion here
    """
        _, u1 = np.split(u, 2)
        return np.dot(self.D2, u1)

    def Jac(self, u, lmbda):
        """
      The jacobian of the system
    """
        jac00 = self.Jac00(u, lmbda)
        jac01 = self.Jac01(u, lmbda)
        jac10 = self.Jac10(u, lmbda)
        jac11 = self.Jac11(u, lmbda)
        return np.block([[jac00, jac01], [jac10, jac11]])

    def Jac00(self, u, lmbda):
        """
        Derivative of Res0 wrt u0
    """
        u0, _ = np.split(u, 2)
        return self.D2 + lmbda * self.Ar * np.diag(
            np.divide(np.exp(np.divide(self.Ar * u0, 1.0 + u0)), np.square(1.0 + u0))
        )

    def Jac01(self, u, lmbda):
        """
        Derivative of Res0 wrt u1
    """
        u0, u1 = np.split(u, 2)
        return np.zeros((len(u0), len(u1)))

    def Jac10(self, u, lmbda):
        """
        Derivative of Res1 wrt u0
    """
        u0, u1 = np.split(u, 2)
        return np.zeros((len(u1), len(u0)))

    def Jac11(self, u, lmbda):
        """
        Derivative of Res1 wrt u1
    """
        return self.D2


def test_coupled():
    problem = Coupled(L=1.0, N=50, nvar=2)
    # Initial guess
    u0 = np.zeros(problem.N + 1)
    u1 = np.zeros(problem.N + 1)
    u = np.hstack((u0, u1))
    # Initial parameter value
    lmbda0 = 0.0

    problem.run(u, lmbda0, 0.1, 800)


if __name__ == "__main__":
    test_coupled()
    plot_continuation_results("test_coupled_out.csv", nvar=2)



#!/usr/bin/env python

import numpy as np

from contspy import Continuation, plot_continuation_results


class Bratu1d(Continuation):
    def Res(self, u, lmbda):
        """
      The residual to solve for the system
    """
        return np.dot(self.D2, u) + lmbda * np.exp(u)

    def dRes_dlmbda(self, u, lmbda):
        """
      The residual derivative wrt the lambda parameter
    """
        return np.exp(u)

    def Jac(self, u, lmbda):
        """
      The jacobian of the system
    """
        return self.D2 + lmbda * np.diag(np.exp(u))


def test_bratu():
    problem = Bratu1d(L=1.0, N=50)
    # Initial guess
    u0 = np.zeros(problem.N + 1)
    # Initial parameter value
    lmbda0 = 0.0

    problem.run(u0, lmbda0, 0.1, 500)


if __name__ == "__main__":
    test_bratu()
    plot_continuation_results("test_bratu_out.csv")



import numpy as np
from termcolor import colored

from .outputs import initialize_output, write_output
from .solvers import solve_Newton
from .spectral import applyBC, cheb


class Continuation:
    def __init__(self, L=1.0, N=20, nvar=1, output_file_base=None):
        # Settings for spectral elements method
        self.L = L  # lengh of the domain
        self.N = N  # N must be even
        self.D, self.D2, self.x = cheb(self.N, self.L)
        # Boundary conditions
        self.D = applyBC(self.D)
        self.D2 = applyBC(self.D2)
        # Number of variables
        self.nvar = nvar

    def Res(self, u, lmbda):
        """
        The residual to solve for the system
    """
        raise Exception("You need to implement this method in your child class!")

    def dRes_dlmbda(self, u, lmbda):
        """
        The residual derivative wrt the lambda parameter
    """
        raise Exception("You need to implement this method in your child class!")

    def Jac(self, u, lmbda):
        """
        The jacobian of the system
    """
        raise Exception("You need to implement this method in your child class!")

    def run(
        self,
        u0,
        lmbda0,
        step_size,
        max_steps,
        abs_tol=1.0e-08,
        rel_tol=1.0e-08,
        max_iters=200,
        filename=None,
        output_steps=False,
    ):
        """
        Pseudo arc length continuation
        INPUT
        u0: initial guess for the solution
        lmbda0: initial guess for the bifurcation parameter
        step_size0: initial arc-length step
        max_steps: the maximum steps to perform
        abs_tol: the absolute tolerance for the Newton solver
        rel_tol: the relative tolerance for the Newton solver
        max_iters: maximum number of iteration for the Newton solver
    """
        # INITIALIZATION
        lmbda = lmbda0  # bifurcation parameter
        k = 0  # iteration
        s = 0  # arc length
        n_saddle = 0
        n_hopf = 0
        output_fname, output_steps_fname = self.initial_step(
            u0, lmbda0, step_size, filename, output_steps
        )

        # # Check regularity and stability of new point
        # eigvals = np.linalg.eigvals(self.Jac(u, lmbda))
        # if np.amax(eigvals.real) == 0:
        #     print("Zero Eigenvalue found!")
        # stability = np.amax(eigvals.real) < 0
        # oscillation = eigvals[np.argmax(eigvals.real)].imag != 0.0
        # saddle = False
        # hopf = False

        # Iteration
        k += 1
        s += step_size
        ds = step_size

        # MAIN LOOP FOR ARC LENGTH
        while True:
            if k > max_steps:
                break

            newton_success = False
            # Loop in case solve fails
            while True:
                if newton_success:
                    break

                print()
                print(f"Step {k}, s = {s:.3e}, ds = {ds:.3e}")

                (
                    newton_success,
                    u,
                    lmbda,
                    du_ds,
                    dlmbda_ds,
                    stability,
                    oscillation,
                    saddle,
                    hopf,
                ) = self.step(ds, abs_tol, rel_tol, max_iters)

                # Cut step size
                if not newton_success:
                    ds /= 2.0

            # Save current calues
            self.u = u
            self.lmbda = lmbda
            self.du_ds = du_ds
            self.dlmbda_ds = dlmbda_ds

            # Output results
            write_output(
                k,
                output_fname,
                output_steps_fname,
                self.x,
                self.u,
                self.lmbda,
                self.nvar,
                stability,
                oscillation,
                saddle,
                hopf,
            )

            # Iteration
            k += 1
            s += ds
            if saddle:
                n_saddle += 1
            if hopf:
                n_hopf += 1

        print()
        print(colored("Arc-length continuation complete!", "green"))
        print(
            "Detected {} saddle points and {} Hopf bifurcation points.".format(
                n_saddle, n_hopf
            )
        )
        return None

    def tangent_predictor(self, u, lmbda, ds):
        """
    """
        # du/dlmbda
        delta = 1.0e-08
        eps1 = delta * (np.abs(lmbda) + delta)
        dRes_dlmbda = (self.Res(u, lmbda + eps1) - self.Res(u, lmbda)) / eps1
        du_dlmbda = np.linalg.solve(self.Jac(u, lmbda), -dRes_dlmbda)

        # dlmbda/ds
        if np.array_equal(u, self.u) and (lmbda == self.lmbda):  # First step
            r = 1.0 if ds > 0.0 else -1.0
        else:
            r = np.dot(du_dlmbda, u - self.u) + (lmbda - self.lmbda)
        dlmbda_ds = np.divide(
            r / np.abs(r), np.sqrt(1.0 + np.dot(du_dlmbda, du_dlmbda))
        )

        # du/ds
        du_ds = du_dlmbda * dlmbda_ds

        return du_ds, dlmbda_ds

    def initial_step(self, u0, lmbda, ds, filename, output_steps):
        """
        Calculate trivial solution with initial bifurcation parameter
        INPUT
        u0: initial guess of the solution
        lmbda: initial bifurcation parameter
        filename: base file name for continuation output
        output_steps: boolean to specify whether each step should be outputted
        OUTPUT
        u: initial solution
        output_fname: output file name (with path) for continuation output
    """
        k0 = 0
        s0 = 0
        ds0 = 0
        print()
        print(f"Step {k0}, s = {s0:.3e}, ds = {ds0:.3e}")

        # Get number of variables
        if len(u0) != (self.nvar * (self.N + 1)):
            raise Exception(
                "Size of the initial guess",
                len(u0),
                "does not match grid resolution and number of variables. Size should be ",
                self.nvar * (self.N + 1),
            )
        u0_vars = np.split(u0, self.nvar)
        for i in range(self.nvar):
            u0_vars[i] = applyBC(u0_vars[i])
        u0 = np.hstack(u0_vars)
        # Get initial solution
        u, _ = solve_Newton(
            lambda u: self.Res(u, lmbda), lambda u: self.Jac(u, lmbda), u0
        )

        # Save in class
        self.lmbda = lmbda
        self.u = u
        self.stability = True

        # Update tangents
        self.du_ds, self.dlmbda_ds = self.tangent_predictor(u, lmbda, ds)

        # Continuation output: initial results
        if self.nvar > 1:
            headers_vars = ["u" + str(int(k)) + "_norm" for k in range(self.nvar)]
            headers_output = [
                "lambda",
                "stability",
                "oscillation",
                "saddle",
                "hopf",
            ]
            headers_output[1:1] = headers_vars
        else:
            headers_output = [
                "lambda",
                "u_norm",
                "stability",
                "oscillation",
                "saddle",
                "hopf",
            ]
        output_fname, output_steps_fname = initialize_output(
            filename, headers_output, output_steps
        )
        write_output(
            0,
            output_fname,
            output_steps_fname,
            self.x,
            self.u,
            self.lmbda,
            self.nvar,
            True,
            False,
            False,
            False,
        )
        print("lmbda = {}, ||x|| = {}".format(self.lmbda, np.linalg.norm(self.u)))

        return output_fname, output_steps_fname

    def step(self, ds, abs_tol, rel_tol, max_iters):
        """
    """
        # Tangent predictor for next values
        u = self.u + self.du_ds * ds
        lmbda = self.lmbda + self.dlmbda_ds * ds

        # Arc length continuation
        # New values of u and lmbda
        u, lmbda, newton_success = self.arc_length_continuation(
            u, lmbda, ds, abs_tol, rel_tol, max_iters,
        )

        # Check regularity and stability of new point
        eigvals, eigvecs = np.linalg.eig(self.Jac(u, lmbda))
        stability = np.amax(eigvals.real) < 0
        oscillation = eigvals[np.argmax(eigvals.real)].imag != 0.0
        saddle = False
        hopf = False
        # Update tangents
        du_ds, dlmbda_ds = self.tangent_predictor(u, lmbda, ds)

        # Check Hopf point
        if self.stability != stability and oscillation:
            print(colored("Hopf point tracking algorithm!", "red"))
            u, lmbda, newton_success = self.hopf_point_locator(eigvals, eigvecs)
            oscillation = True
            saddle = False
            hopf = True

            # # Update tangents
            # du_ds, dlmbda_ds = self.tangent_predictor(u, lmbda, ds)

        # Check saddle point
        elif self.dlmbda_ds * dlmbda_ds < 0.0:
            print(colored("Saddle point tracking algorithm!", "red"))
            u, lmbda, newton_success = self.saddle_point_locator()

            stability = False
            oscillation = False
            saddle = True
            hopf = False

            # # Update tangents
            # du_ds, dlmbda_ds = self.tangent_predictor(u, lmbda, ds)

        return (
            newton_success,
            u,
            lmbda,
            du_ds,
            dlmbda_ds,
            stability,
            oscillation,
            saddle,
            hopf,
        )

    def arc_length_continuation(
        self, u, lmbda, ds, abs_tol, rel_tol, max_iters,
    ):
        """
        Arc length continuation. Implementation based on:
        http://www.cs.sandia.gov/LOCA/loca1.1_book.pdf
        INPUT

        OUTPUT
    """
        # Iterations
        num_newton_steps = 0
        newton_success = False

        # Residuals and Jacobian
        J = self.Jac(u, lmbda)
        Ru = self.Res(u, lmbda)
        Rlmbda = (
            np.dot(self.du_ds, u - self.u) + (lmbda - self.lmbda) * self.dlmbda_ds - ds
        )

        Rnrm = np.sqrt(np.linalg.norm(Ru) ** 2 + Rlmbda ** 2)
        Rnrm0 = Rnrm

        print(f"{num_newton_steps} Nonlinear |R| = ", colored(f"{Rnrm:.6e}", "green"))

        while num_newton_steps < max_iters:
            # Convergence criteria
            if Rnrm < abs_tol or np.divide(Rnrm, Rnrm0) < rel_tol:
                if Rnrm < abs_tol:
                    reason = "ABS_TOL"
                else:
                    reason = "REL_TOL"
                print(
                    f"Solve converged due to {reason} after {num_newton_steps} iterations."
                )
                print("lmbda = {}, ||x|| = {}".format(lmbda, np.linalg.norm(u)))
                print(colored("Solve Converged!", "green"))
                newton_success = True
                return u, lmbda, newton_success

            # Update solution
            # Solve
            #
            #  (J,     dR/dlmbda) (du    )  =  -(Ru)
            #  (du/ds, dlmbda/ds) (dlmbda)      (Rlmbda)
            #

            # dR/dlmbda
            delta = 1.0e-06
            eps1 = delta * (np.abs(lmbda) + delta)
            Resb = (self.Res(u, lmbda + eps1) - Ru) / eps1

            # Temporary vectors
            a = np.linalg.solve(J, -Ru)
            b = np.linalg.solve(J, -Resb)

            # Update values
            dlmbda = -np.divide(
                Rlmbda + np.dot(self.du_ds, a), self.dlmbda_ds + np.dot(self.du_ds, b)
            )
            du = a + dlmbda * b

            u += du
            lmbda += dlmbda
            num_newton_steps += 1

            # Update residuals
            Ru = self.Res(u, lmbda)
            Rlmbda = (
                np.dot(self.du_ds, u - self.u)
                + (lmbda - self.lmbda) * self.dlmbda_ds
                - ds
            )
            # Update jacobien
            J = self.Jac(u, lmbda)

            Rnrm_old = Rnrm
            Rnrm = np.sqrt(np.linalg.norm(Ru) ** 2 + Rlmbda ** 2)
            if Rnrm < 0.8 * Rnrm_old:
                color = "green"
            elif Rnrm > Rnrm_old:
                color = "red"
            else:
                color = "yellow"

            print(f"{num_newton_steps} Nonlinear |R| = ", colored(f"{Rnrm:.6e}", color))

        # Here, no solutions have been found after max_iters iterations
        print("Solve diverged due to MAX_ITER.")
        print("lmbda = {}, ||x|| = {}".format(lmbda, np.linalg.norm(u)))
        print(colored("Solve did NOT converge!", "red"))
        return u, lmbda, newton_success

    def saddle_point_locator(self, abs_tol=1.0e-04, rel_tol=1.0e-04, max_iters=1000):
        # Initialization
        u = self.u.copy()
        lmbda = self.lmbda.copy()

        # Iterations
        num_newton_steps = 0
        newton_success = False

        # Residuals and Jacobian
        J = self.Jac(u, lmbda)
        Ru = self.Res(u, lmbda)
        # dR/dlmbda
        delta = 1.0e-06
        eps1 = delta * (np.abs(lmbda) + delta)
        Resb = (self.Res(u, lmbda + eps1) - Ru) / eps1
        b = np.linalg.solve(J, -Resb)
        v = b / np.linalg.norm(b)
        phi = b / np.linalg.norm(b)
        Rv = np.dot(J, v)
        Rlmbda = np.dot(phi, v) - 1.0

        Rnrm = np.sqrt(np.linalg.norm(Ru) ** 2 + np.linalg.norm(Rv) ** 2 + Rlmbda ** 2)
        Rnrm0 = Rnrm

        print(f"{num_newton_steps} Nonlinear |R| = ", colored(f"{Rnrm:.6e}", "green"))

        while num_newton_steps < max_iters:
            # Convergence criteria
            if Rnrm < abs_tol or np.divide(Rnrm, Rnrm0) < rel_tol:
                if Rnrm < abs_tol:
                    reason = "ABS_TOL"
                else:
                    reason = "REL_TOL"
                print(
                    f"Solve converged due to {reason} after {num_newton_steps} iterations."
                )
                print("lmbda = {}, ||x|| = {}".format(lmbda, np.linalg.norm(u)))
                print(colored("Solve Converged!", "green"))
                newton_success = True
                return u, lmbda, newton_success

            # Update solution
            # Solve
            #
            #  (J,      0,   dR/dlmbda )  (du)       -(Ru)
            #  (dJv/dx, J,   dJv/dlmbda)  (dv)     = -(Jv)
            #  (0,      phi, 0         )  (dlmbda)   -(phi v - 1)
            #

            # Temporary vectors
            # Estimate jacobian with 1st order finite differences
            delta = 1.0e-06
            a = np.linalg.solve(J, -Ru)

            eps1 = delta * (np.abs(lmbda) + delta)
            Resb = (self.Res(u, lmbda + eps1) - Ru) / eps1
            b = np.linalg.solve(J, -Resb)

            eps2 = delta * (np.linalg.norm(u) / np.linalg.norm(a) + delta)
            Resc = (np.dot(self.Jac(u + eps2 * a, lmbda), v) - np.dot(J, v)) / eps2
            c = np.linalg.solve(J, -Resc)

            eps3 = delta * (np.linalg.norm(u) / np.linalg.norm(b) + delta)
            Resd = (
                np.dot(self.Jac(u + eps3 * b, lmbda), v) / eps3
                + np.dot(self.Jac(u, lmbda + eps1), v) / eps1
                - np.dot(J, v) * (1.0 / eps3 + 1.0 / eps1)
            )
            d = np.linalg.solve(J, -Resd)

            # Update values
            dlmbda = (1.0 - np.dot(phi, c)) / np.dot(phi, d)
            du = a + dlmbda * b
            dv = c + dlmbda * d - v

            u += du
            v += dv
            lmbda += dlmbda
            num_newton_steps += 1

            # Update residuals and Jacobian
            J = self.Jac(u, lmbda)
            Ru = self.Res(u, lmbda)
            Rv = np.dot(J, v)
            Rlmbda = np.dot(phi, v) - 1.0

            Rnrm_old = Rnrm
            Rnrm = np.sqrt(
                np.linalg.norm(Ru) ** 2 + np.linalg.norm(Rv) ** 2 + Rlmbda ** 2
            )
            if Rnrm < 0.8 * Rnrm_old:
                color = "green"
            elif Rnrm > Rnrm_old:
                color = "red"
            else:
                color = "yellow"

            print(f"{num_newton_steps} Nonlinear |R| = ", colored(f"{Rnrm:.6e}", color))

        # Here, no solutions have been found after max_iters iterations
        print("Solve diverged due to MAX_ITER.")
        print("lmbda = {}, ||x|| = {}".format(lmbda, np.linalg.norm(u)))
        print(colored("Solve did NOT converge!", "red"))
        return u, lmbda, newton_success

    def hopf_point_locator(
        self,
        eigenvals_current,
        eigenvecs_current,
        abs_tol=1.0e-04,
        rel_tol=1.0e-04,
        max_iters=1000,
    ):
        # Initialization
        u = self.u.copy()
        lmbda = self.lmbda.copy()

        # Iterations
        num_newton_steps = 0
        newton_success = False

        # Residuals and Jacobian
        J = self.Jac(u, lmbda)
        ind = np.argmax(eigenvals_current.real)
        Ru = self.Res(u, lmbda)
        # dR/dlmbda
        delta = 1.0e-06
        eps1 = delta * (np.abs(lmbda) + delta)
        Resb = (self.Res(u, lmbda + eps1) - Ru) / eps1
        b = np.linalg.solve(J, -Resb)
        v = eigenvecs_current[:, ind].real
        phi = b / np.linalg.norm(b)
        w = eigenvecs_current[:, ind].imag
        omega = eigenvals_current[ind].imag
        Rv = (
            np.dot(J, v) + omega * w
        )  # here I assume that B is the identity matrix (neglect thermal pressurization)
        Rw = (
            np.dot(J, w) - omega * v
        )  # here I assume that B is the identity matrix (neglect thermal pressurization)
        Romega = np.dot(phi, v) - 1.0
        Rlmbda = np.dot(phi, w)

        Rnrm = np.sqrt(
            np.linalg.norm(Ru) ** 2
            + np.linalg.norm(Rv) ** 2
            + np.linalg.norm(Rw) ** 2
            + Romega ** 2
            + Rlmbda ** 2
        )
        Rnrm0 = Rnrm

        print(f"{num_newton_steps} Nonlinear |R| = ", colored(f"{Rnrm:.6e}", "green"))

        while num_newton_steps < max_iters:
            # Convergence criteria
            if Rnrm < abs_tol or np.divide(Rnrm, Rnrm0) < rel_tol:
                if Rnrm < abs_tol:
                    reason = "ABS_TOL"
                else:
                    reason = "REL_TOL"
                print(
                    f"Solve converged due to {reason} after {num_newton_steps} iterations."
                )
                print("lmbda = {}, ||x|| = {}".format(lmbda, np.linalg.norm(u)))
                print(colored("Solve Converged!", "green"))
                newton_success = True
                return u, lmbda, newton_success

            # Update solution
            # Solve
            #
            #  (J,      0,        0,       0,    dR/dlmbda )  (du)       -(Ru)
            #  (dJv/du, J,        omega B, B w,  dJv/dlmbda)  (dv)       -(Jv + omega B w)
            #  (dJw/du, -omega B, J,       -B v, dJw/dlmbda)  (dw)     = -(Jw - omega B v)
            #  (0,      phi,      0,       0,    0         )  (domega)   -(phi v - 1)
            #  (0,      0,        phi,     0,    0         )  (dlmbda)   -(phi w)
            #

            # Temporary vectors
            # Estimate jacobian with 1st order finite differences
            delta = 1.0e-06
            a = np.linalg.solve(J, -Ru)

            eps1 = delta * (np.abs(lmbda) + delta)
            Resb = (self.Res(u, lmbda + eps1) - Ru) / eps1
            b = np.linalg.solve(J, -Resb)

            J_hopf = np.block(
                [[J, omega * np.identity(Ru.size)], [-omega * np.identity(Ru.size), J]]
            )
            Rescd = np.concatenate((-w, v))
            sol_cd = np.linalg.solve(J_hopf, -Rescd)
            sol_cd = np.split(sol_cd, 2)
            c = sol_cd[0]
            d = sol_cd[1]

            eps2 = delta * (np.linalg.norm(u) / np.linalg.norm(a) + delta)
            Rese = (np.dot(self.Jac(u + eps2 * a, lmbda), v) - np.dot(J, v)) / eps2
            Resf = (np.dot(self.Jac(u + eps2 * a, lmbda), w) - np.dot(J, w)) / eps2
            Resef = np.concatenate((Rese, Resf))
            sol_ef = np.linalg.solve(J_hopf, -Resef)
            sol_ef = np.split(sol_ef, 2)
            e = sol_ef[0]
            f = sol_ef[1]

            eps3 = delta * (np.linalg.norm(u) / np.linalg.norm(b) + delta)
            Resg = (
                np.dot(self.Jac(u + eps3 * b, lmbda), v) / eps3
                + np.dot(self.Jac(u, lmbda + eps1), v) / eps1
                - np.dot(J, v) * (1.0 / eps3 + 1.0 / eps1)
            )
            Resh = (
                np.dot(self.Jac(u + eps3 * b, lmbda), w) / eps3
                + np.dot(self.Jac(u, lmbda + eps1), w) / eps1
                - np.dot(J, w) * (1.0 / eps3 + 1.0 / eps1)
            )
            Resgh = np.concatenate((Resg, Resh))
            sol_gh = np.linalg.solve(J_hopf, -Resgh)
            sol_gh = np.split(sol_gh, 2)
            g = sol_gh[0]
            h = sol_gh[1]

            # Update values
            dlmbda = (
                np.dot(phi, c) * np.dot(phi, f)
                - np.dot(phi, e) * np.dot(phi, d)
                + np.dot(phi, d)
            ) / (np.dot(phi, d) * np.dot(phi, g) - np.dot(phi, c) * np.dot(phi, h))
            domega = (dlmbda * np.dot(phi, h) + np.dot(phi, f)) / np.dot(phi, d)
            du = a + dlmbda * b
            dv = e + dlmbda * g - domega * c - v
            dw = f + dlmbda * h - domega * d - w

            u += du
            v += dv
            w += dw
            omega += domega
            lmbda += dlmbda
            num_newton_steps += 1

            # Update residuals and jacobian
            J = self.Jac(u, lmbda)
            Ru = self.Res(u, lmbda)
            Rv = (
                np.dot(J, v) + omega * w
            )  # here I assume that B is the identity matrix (neglect thermal pressurization)
            Rw = (
                np.dot(J, w) - omega * v
            )  # here I assume that B is the identity matrix (neglect thermal pressurization)
            Romega = np.dot(phi, v) - 1.0
            Rlmbda = np.dot(phi, w)

            Rnrm_old = Rnrm
            Rnrm = np.sqrt(
                np.linalg.norm(Ru) ** 2
                + np.linalg.norm(Rv) ** 2
                + np.linalg.norm(Rw) ** 2
                + Romega ** 2
                + Rlmbda ** 2
            )
            if Rnrm < 0.8 * Rnrm_old:
                color = "green"
            elif Rnrm > Rnrm_old:
                color = "red"
            else:
                color = "yellow"

            print(f"{num_newton_steps} Nonlinear |R| = ", colored(f"{Rnrm:.6e}", color))

        # Here, no solutions have been found after max_iters iterations
        print("Solve diverged due to MAX_ITER.")
        print("lmbda = {}, ||x|| = {}".format(lmbda, np.linalg.norm(u)))
        print(colored("Solve did NOT converge!", "red"))
        return u, lmbda, newton_success



import sys

from .__about__ import __version__


def show(argv=None):
    # Parse command line arguments.
    parser = _get_parser()
    args = parser.parse_args(argv)

    print(args.number)
    return


def _get_parser():
    import argparse

    parser = argparse.ArgumentParser(
        description=("Dummy contspy executable."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("number", type=int, help="number to show")

    __copyright__ = "Copyright (c) 2020 Antoine Jacquey <ajacquey@mit.edu>"
    version_text = "\n".join(
        [
            "contspy {} [Python {}.{}.{}]".format(
                __version__,
                sys.version_info.major,
                sys.version_info.minor,
                sys.version_info.micro,
            ),
            __copyright__,
        ]
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=version_text,
        help="display version information",
    )

    return parser



import numpy as np


def cheb(N, L=1.0):
    """
    Chebyshev polynomial differentiation matrix.
    Ref.: Trefethen's 'Spectral Methods in MATLAB' book.
    INPUT
    N: discretization points number (must be even)
    L: size of the domain (default is 1)
    OUTPUT
    D: Differentiation matrix
    D2: Differentiation matrix second order
    x: Spatial discretization
  """
    assert N % 2 == 0, "N should be even!"  # only when N is even!
    x = L / 2.0 * np.cos(np.pi * np.arange(0, N + 1) / N)
    x[int(N / 2)] = 0.0
    c = np.ones(N + 1)
    c[0] = 2.0
    c[N] = 2.0
    c = c * (-1.0) ** np.arange(0, N + 1)
    c = c.reshape(N + 1, 1)
    X = np.tile(x.reshape(N + 1, 1), (1, N + 1))
    dX = X - X.T
    D = np.dot(c, 1.0 / c.T) / (dX + np.eye(N + 1))
    D = D - np.diag(D.sum(axis=1))
    D2 = np.dot(D, D)
    return D, D2, x


def applyBC(array):
    """
    Apply boundary conditions
    INPUT:
    D: Differentiation matrix
    D2: Differentiation matrix second order
    x: Spatial discretization
    OUTPUT:
    D: Differentiation matrix
    D2: Differentiation matrix second order
    x: Spatial discretization
  """
    arr_shape = array.shape
    if len(arr_shape) == 1:
        N = len(array) - 1
        return array[1:N]
    elif len(arr_shape) == 2:
        N = len(array[0, :]) - 1
        return array[1:N, 1:N]
    else:
        raise Exception("Cannot apply BC, invalid ipunt array!")



#!/usr/bin/env python

import numpy as np

from contspy import Continuation, plot_continuation_results


class SCurve(Continuation):
    def __init__(self, L, N):
        Continuation.__init__(self, L, N)
        self.Ar = 5.0

    def Res(self, u, lmbda):
        """
      The residual to solve for the system
    """
        return np.dot(self.D2, u) + lmbda * np.exp(np.divide(self.Ar * u, 1.0 + u))

    def Jac(self, u, lmbda):
        """
      The jacobian of the system
    """
        return self.D2 + lmbda * self.Ar * np.diag(
            np.divide(np.exp(np.divide(self.Ar * u, 1.0 + u)), np.square(1.0 + u))
        )


def test_s_curve():
    problem = SCurve(L=1.0, N=50)
    # Initial guess
    u0 = np.zeros(problem.N + 1)
    # Initial parameter value
    lmbda0 = 0.0

    problem.run(u0, lmbda0, 0.1, 800)


if __name__ == "__main__":
    test_s_curve()
    plot_continuation_results("test_S_curve_out.csv")



import os
import sys

import matplotlib.pyplot as plt
import numpy as np


def split_stability_results(u_norm, stability, oscillation):
    """

"""
    u_norm_stable_reg = np.ones_like(u_norm) * np.NaN
    u_norm_unstable_reg = np.ones_like(u_norm) * np.NaN
    u_norm_stable_osc = np.ones_like(u_norm) * np.NaN
    u_norm_unstable_osc = np.ones_like(u_norm) * np.NaN

    u_norm_stable_reg[(stability) & ~oscillation] = u_norm[(stability) & ~oscillation]
    u_norm_unstable_reg[~stability & ~oscillation] = u_norm[~stability & ~oscillation]
    u_norm_stable_osc[stability & oscillation] = u_norm[stability & oscillation]
    u_norm_unstable_osc[~stability & oscillation] = u_norm[~stability & oscillation]

    return (
        u_norm_stable_reg,
        u_norm_stable_osc,
        u_norm_unstable_reg,
        u_norm_unstable_osc,
    )


def get_bifurcation_points(lmbda, u_norm, saddle, hopf):
    """

"""
    lmbda_saddle = lmbda[saddle]
    u_norm_saddle = u_norm[saddle]
    lmbda_hopf = lmbda[hopf]
    u_norm_hopf = u_norm[hopf]

    return lmbda_saddle, lmbda_hopf, u_norm_saddle, u_norm_hopf


def plot_continuation_results(filename, nvar=1):
    """

"""
    # Path to script
    script_file = os.path.abspath(sys.argv[0]).split("/")[-1]
    filename = os.path.abspath(sys.argv[0]).replace(script_file, filename)
    # Check if file exists
    if not os.path.exists(filename):
        raise Exception(
            "File",
            filename,
            "does not exits, please check the name and path of the given file!",
        )

    print()
    print("Reading data stored in file", filename, "...")

    # Load data
    cols = [k + 1 for k in range(nvar)]
    lmbda = np.loadtxt(
        filename, dtype=float, delimiter=",", skiprows=1, unpack=True, usecols=[0]
    )
    u_norm = np.loadtxt(
        filename, dtype=float, delimiter=",", skiprows=1, unpack=False, usecols=cols
    )
    cols = [k + nvar + 1 for k in range(4)]
    stability, oscillation, saddle, hopf = np.loadtxt(
        filename, dtype=bool, delimiter=",", skiprows=1, unpack=True, usecols=cols,
    )

    if nvar > 1:
        u_norm = list(u_norm.T)

        u_norm_stable_reg = [None] * nvar
        u_norm_stable_osc = [None] * nvar
        u_norm_unstable_reg = [None] * nvar
        u_norm_unstable_osc = [None] * nvar
        u_norm_saddle = [None] * nvar
        u_norm_hopf = [None] * nvar

        for k in range(nvar):
            (
                u_norm_stable_reg[k],
                u_norm_stable_osc[k],
                u_norm_unstable_reg[k],
                u_norm_unstable_osc[k],
            ) = split_stability_results(u_norm[k], stability, oscillation)
            (
                lmbda_saddle,
                lmbda_hopf,
                u_norm_saddle[k],
                u_norm_hopf[k],
            ) = get_bifurcation_points(lmbda, u_norm[k], saddle, hopf)
    else:
        (
            u_norm_stable_reg,
            u_norm_stable_osc,
            u_norm_unstable_reg,
            u_norm_unstable_osc,
        ) = split_stability_results(u_norm, stability, oscillation)
        lmbda_saddle, lmbda_hopf, u_norm_saddle, u_norm_hopf = get_bifurcation_points(
            lmbda, u_norm, saddle, hopf
        )

    # Figure name
    fig_filename = os.path.splitext(os.path.abspath(filename))[0] + ".png"

    # Plot
    plt.rc("text", usetex=True)
    fig, axes = plt.subplots(1, nvar)
    if nvar > 1:
        for k in range(nvar):
            initialize_plot(axes[k], r"$||u_{" + str(k) + r"}||_{\infty}$")
            plot_continuation_lines(
                axes[k],
                lmbda,
                u_norm_stable_reg[k],
                u_norm_stable_osc[k],
                u_norm_unstable_reg[k],
                u_norm_unstable_osc[k],
            )
            plot_continuation_points(
                axes[k], lmbda_saddle, u_norm_saddle[k], lmbda_hopf, u_norm_hopf[k]
            )
    else:
        initialize_plot(axes, r"$||u||_{\infty}$")
        plot_continuation_lines(
            axes,
            lmbda,
            u_norm_stable_reg,
            u_norm_stable_osc,
            u_norm_unstable_reg,
            u_norm_unstable_osc,
        )
        plot_continuation_points(
            axes, lmbda_saddle, u_norm_saddle, lmbda_hopf, u_norm_hopf
        )

    # ax.set_xlim(0.0, 1.2)
    # ax.set_ylim(0.0, 15.0)
    fig.set_size_inches(8 * nvar, 6)
    fig.savefig(fig_filename, type="PNG", bbox_inches="tight", transparent=False)
    print("Plot of continuation results in file", fig_filename, "completed!")

    return None


def initialize_plot(ax, label):
    """

"""
    ax.set_xlabel(r"$\lambda$", fontsize=18)
    ax.set_ylabel(label, fontsize=18)
    ax.tick_params(axis="x", labelsize=16)
    ax.tick_params(axis="y", labelsize=16)
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.grid()
    return None


def plot_continuation_lines(
    ax, lmbda, u_stable_reg, u_stable_osc, u_unstable_reg, u_unstable_osc
):
    """
    Plot the stable/unstable and regular/oscillatory solutions for the contiuation
    INPUT:
    ax: matplotlib axes
    lmbda: array containing the bifurcation parameter values
    u_stable_reg: infinite norm of the variable for stable and regular solutions
    u_stable_osc: infinite norm of the variable for stable and oscillatory solutions
    u_unstable_reg: infinite norm of the variable for unstable and regular solutions
    u_unstable_osc: infinite norm of the variable for unstable and oscillatory solutions
"""
    ax.plot(lmbda, u_stable_reg, "-", color="blue")
    ax.plot(lmbda, u_unstable_reg, "--", color="blue")
    ax.plot(lmbda, u_stable_osc, "-", color="red")
    ax.plot(lmbda, u_unstable_osc, "--", color="red")

    return None


def plot_continuation_points(ax, lmbda_saddle, u_saddle, lmbda_hopf, u_hopf):
    """
    Plot the saddle and Hopf bifurcation points from the continuation results
    INPUT
    ax: matplotlib axes
    lmbda_saddle: array of the bifurcation parameter values corresponding to saddle points
    u_saddle: array of the infinite norms of the variable corresponding to saddle points
    lmbda_hopf: array of the bifurcation parameter values corresponding to Hopf points
    u_hopf: array of the infinite norms of the variable corresponding to Hopf points
"""
    ax.plot(lmbda_saddle, u_saddle, "o", color="blue")
    ax.plot(lmbda_hopf, u_hopf, "o", color="red")

    return None



import numpy as np


def solve_Newton(
    R, J, x0, abs_tol=1.0e-10, rel_tol=1.0e-10, max_iters=200, debug=False
):
    """
    Solve nonlinear system R=0 by Newton's method.
    J is the Jacobian of R. Both R and J must be functions of x.
    At input, x holds the start value. The iteration continues
    until ||F|| < abs_tol or ||F|| / ||F0|| < rel_tol.
  """

    # Initial Residuals
    x = x0
    Rx = R(x)
    Rnrm = np.linalg.norm(Rx)
    Rnrm0 = Rnrm

    # Iteration
    k = 0
    while k < max_iters:
        # Convergence criteria
        if Rnrm < abs_tol or np.divide(Rnrm, Rnrm0) < rel_tol:
            return x, k

        # Update solution
        dx = np.linalg.solve(J(x), -Rx)
        x += dx
        Rx = R(x)
        Rnrm = np.linalg.norm(Rx)  # l2 norm of vector
        k += 1

    # Here, no solutions have been found after max_iters iterations
    raise Exception(
        "No solution has been found after ", k, " iterations in the Newton solve."
    )



import os
import sys

import numpy as np


def initialize_output(filename, headers, output_steps):
    """
    Initialize the output of continuation in a CSV file
"""
    output_fname = get_output_filename(filename)
    np.savetxt(
        output_fname, [np.asarray(headers)], delimiter=",", fmt="%s", comments=""
    )

    if output_steps:
        # Here we need to create the folder 'steps' if it does not exist
        csv_filename = output_fname.split("/")[-1]
        path = output_fname.replace(csv_filename, "steps/")
        isdir = os.path.isdir(path)

        if not isdir:
            try:
                os.mkdir(path)
            except OSError:
                print("Creation of the directory %s failed" % path)

        output_steps_fname = path + ".".join(csv_filename.split(".")[:-1]) + "_"
    else:
        output_steps_fname = ""

    return output_fname, output_steps_fname


def write_output(
    k,
    output_fname,
    output_steps_fname,
    x,
    u,
    lmbda,
    nvar,
    stability,
    oscillation,
    saddle,
    hopf,
):
    """
    Output continuation step and spectral step in a CSV file
"""
    # Continuation
    if nvar > 1:
        uvars = np.split(u, nvar)
        results_u = [np.linalg.norm(uvar, np.inf) for uvar in uvars]
        results = [
            lmbda,
            int(stability),
            int(oscillation),
            int(saddle),
            int(hopf),
        ]
        results[1:1] = results_u
    else:
        results = [
            lmbda,
            np.linalg.norm(u, np.inf),
            int(stability),
            int(oscillation),
            int(saddle),
            int(hopf),
        ]

    fmt = ["%1.4e", "%1i", "%1i", "%1i", "%1i"]
    fmt_var = ["%1.4e"] * nvar
    fmt[1:1] = fmt_var
    with open(output_fname, "a+", newline="") as write_obj:
        np.savetxt(
            write_obj, [results], fmt=fmt, comments="", delimiter=",",
        )

    # Spectral
    if bool(output_steps_fname):  # string not empty
        filename = output_steps_fname + str(k) + ".csv"
        if nvar > 1:
            uvars = np.split(u, nvar)
            uvars = [np.concatenate([[0.0], uvar, [0.0]]) for uvar in uvars]
            uvars = np.array(uvars)
            header = ["x"]
            header_var = ["u" + str(int(k)) for k in range(nvar)]
            header[1:1] = header_var
            header = ",".join(header)
            data = np.column_stack((np.flip(x), uvars.transpose()))
        else:
            u = np.concatenate([[0.0], u, [0.0]])
            header = "x,u"
            data = np.column_stack((np.flip(x), u))
        np.savetxt(
            filename, data, delimiter=",", fmt="%1.4e", header=header, comments=""
        )
    return None


def get_output_filename(filename):
    """
  Get the file name for the output CSV file
"""
    if filename is None:
        script_fname = os.path.splitext(os.path.abspath(sys.argv[0]))[0]
        output_fname = script_fname + "_out.csv"
    else:
        path_to_script = os.path.abspath(os.path.dirname(sys.argv[0]))
        output_fname = path_to_script + "/" + filename + "_out.csv"

    return output_fname



from .__about__ import __version__
from .main import Continuation
from .plotting import plot_continuation_results

__all__ = [
    "__version__",
    "Continuation",
    "plot_continuation_results",
]
